from rust_circuit import (
    Add as rAdd,
    ScalarConstant as rScalarConstant,
    Rearrange as rRearrange,
    RearrangeSpec as rRearrangeSpec,
    ArrayConstant as rArrayConstant,
    compiler_simp,
    compiler_simp_until_same,
    strip_names,
    deep_canonicalize,
    deep_maybe_distribute,
    deep_optimize_einsums,
    optimize_circuit,
    count_nodes as r_count_nodes,
    scheduled_evaluate,
    optimize_and_evaluate,
    OptimizationSettings,
    batch_inputs_axis_len,
    batch_largest,
    deep_heuristic_nest_adds,
)
from interp.circuit.circuit import MemoizedFn
from interp.circuit.circuit_compiler.compiler import evaluate_circuit
import torch
from interp.circuit.circuit_compiler.canonical_ast import (
    ast_count_nodes,
    circuit_to_canonical_ast,
    optimize_with_backreferences,
    print_ast_deep,
)
from interp.circuit.circuit_compiler.util import FrozenDict, oom_fmt
from interp.circuit.circuit_utils import cast_circuit, count_nodes, evaluate_fn
from interp.circuit.computational_node import (
    Add,
    Concat,
    Einsum,
    GeneralFunction,
    Index,
    UnaryRearrange,
    log_softmax_fn,
    relu,
    sigmoid,
    sigmoid_fn,
    softmax,
    softmax_fn,
    rsqrt_fn,
    softmax_jacobian,
    reciprocal_jacobian,
    log_exp_p_1_fn,
    sigmoid_jacobian,
    log_exp_p_1_jacobian,
    rsqrt_jacobian,
    reciprocal_fn,
)  # imports used in eval
from interp.circuit.var import AutoTag
from interp.circuit.constant import ArrayConstant, Constant, FloatConstant, One, Zero
from interp.circuit.interop_rust.interop_rust import py_to_rust, rust_to_py
from uuid import UUID
from interp.tools.perf_timer import catchtime

x1 = rRearrange(
    rAdd(
        [
            rArrayConstant(torch.randn((2,))),
            rRearrange(
                rRearrange(rScalarConstant(1.0, ()), rRearrangeSpec([], [[0]], [2])),
                rRearrangeSpec([[0]], [[0], [1]], [None, 2]),
            ),
        ]
    ),
    rRearrangeSpec([[0], [1]], [[0], [1], [2]], [None, None, 2]),
)
x1_s = compiler_simp(deep_canonicalize(x1))
x1_s.compiler_print()
nb_example = eval(
    """
(lambda c17: (lambda c16: (lambda c14: (lambda c15: (lambda c11: (lambda c10: (lambda c13: (lambda c12: (lambda c9: (lambda c2: (lambda c8: (lambda c1: (lambda c7: (lambda c3: (lambda c4: (lambda c0: (lambda c6: (lambda c5: Add(items=FrozenDict({Einsum(args=((Einsum(args=((Einsum(args=((c3, (0,)), (Add(items=FrozenDict({UnaryRearrange(node=Einsum(args=((c6, (0,)), (Add(items=FrozenDict({UnaryRearrange(node=One(shape=(8192,), name='one_sample'), op_string='b0 -> b0', axes_lengths=(), name='one_sample_rearrange_for_add_0', shape=(8192,), is_constant=True, is_explicitly_computable=True, can_be_sampled=True): 1.0, UnaryRearrange(node=c4, op_string='b0 -> b0', axes_lengths=(), name='is_p_var_sample_rearrange_for_add_1', shape=(8192,), is_constant=True, is_explicitly_computable=True, can_be_sampled=True): -1.0}), name='is_q_var_sample', shape=(8192,), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), (0,))), out_axes=(0,), name='logit_diff_times_correct_sample', shape=(8192,), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), op_string='b0 -> b0', axes_lengths=(), name='logit_diff_times_correct_sample_rearrange_for_add_0', shape=(8192,), is_constant=True, is_explicitly_computable=True, can_be_sampled=True): 1.0, UnaryRearrange(node=GeneralFunction(node=c6, function=log_exp_p_1_fn, get_jacobian=log_exp_p_1_jacobian, name='logit_diff_log_exp_p_1_sample', allows_batching=True, non_batch_dims=(), shape=(8192,), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), op_string='b0 -> b0', axes_lengths=(), name='logit_diff_log_exp_p_1_sample_rearrange_for_add_1', shape=(8192,), is_constant=True, is_explicitly_computable=True, can_be_sampled=True): -1.0}), name='log_loss_sample', shape=(8192,), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), (0,))), out_axes=(), name='log_loss_sample_sample_expectation', shape=(), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), ()), (c1, ())), out_axes=(), name='ScalarMul', shape=(), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), ()),), out_axes=(), name='Ein', shape=(), is_constant=True, is_explicitly_computable=True, can_be_sampled=True): 1.0, Einsum(args=((UnaryRearrange(node=Einsum(args=((Einsum(args=((c3, (0,)), (GeneralFunction(node=c6, function=sigmoid_fn, get_jacobian=sigmoid_jacobian, name='logit_diff_sigmoid_sample', allows_batching=True, non_batch_dims=(), shape=(8192,), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), (0,))), out_axes=(), name='logit_diff_sigmoid_sample_sample_expectation', shape=(), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), ()), (c1, ())), out_axes=(), name='ScalarMul', shape=(), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), op_string=' -> ', axes_lengths=(), name='I k1 logit_diff_sigmoid I perm', shape=(), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), ()),), out_axes=(), name='Ein', shape=(), is_constant=True, is_explicitly_computable=True, can_be_sampled=True): 1.0, Einsum(args=((Einsum(args=((Einsum(args=((Add(items=FrozenDict({Einsum(args=((c3, (0,)), (Einsum(args=((c5, (0,)), (c5, (0,))), out_axes=(0,), name='outer_sample', shape=(8192,), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), (0,))), out_axes=(), name='outer_sample_sample_expectation', shape=(), is_constant=True, is_explicitly_computable=True, can_be_sampled=True): 1.0}), name='is_it_p_variance_decompose', shape=(), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), ()), (c1, ())), out_axes=(), name='ScalarMul', shape=(), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), ()), (c1, ())), out_axes=(), name='ScalarMul', shape=(), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), ()),), out_axes=(), name='Ein', shape=(), is_constant=True, is_explicitly_computable=True, can_be_sampled=True): 1.0, Einsum(args=((Einsum(args=((c0, ()), (c1, ())), out_axes=(), name='ScalarMul', shape=(), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), ()),), out_axes=(), name='Ein', shape=(), is_constant=True, is_explicitly_computable=True, can_be_sampled=True): 1.0}), name='Add', shape=(), is_constant=True, is_explicitly_computable=True, can_be_sampled=True))(Add(items=FrozenDict({UnaryRearrange(node=c4, op_string='b0 -> b0', axes_lengths=(), name='is_p_var_sample_rearrange_for_add_0', shape=(8192,), is_constant=True, is_explicitly_computable=True, can_be_sampled=True): 1.0, UnaryRearrange(node=UnaryRearrange(node=c0, op_string=' -> 8192', axes_lengths=(), name='is_p_var_sample_sample_expectation_sample', shape=(8192,), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), op_string='b0 -> b0', axes_lengths=(), name='is_p_var_sample_sample_expectation_sample_rearrange_for_add_1', shape=(8192,), is_constant=True, is_explicitly_computable=True, can_be_sampled=True): -1.0}), name='is_p_var_centered_sample', shape=(8192,), is_constant=True, is_explicitly_computable=True, can_be_sampled=True)))(Index(node=Add(items=FrozenDict({UnaryRearrange(node=Einsum(args=((c7, (0, 1, 2)), (UnaryRearrange(node=Index(node=c8, index=(30, slice(None, None, None)), name='w.unembed_q_mark', hash_tensor_idx_by_value=False, shape=(256,), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), op_string='a -> 8192 a', axes_lengths=(), name='w.unembed_q_mark_sample', shape=(8192, 256), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), (0, 2))), out_axes=(0, 1), name='logits_q_mark_sample', shape=(8192, 35), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), op_string='b0 a0 -> b0 a0', axes_lengths=(), name='logits_q_mark_sample_rearrange_for_add_0', shape=(8192, 35), is_constant=True, is_explicitly_computable=True, can_be_sampled=True): 1.0, UnaryRearrange(node=Einsum(args=((c7, (0, 1, 2)), (UnaryRearrange(node=Index(node=c8, index=(13, slice(None, None, None)), name='w.unembed_period', hash_tensor_idx_by_value=False, shape=(256,), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), op_string='a -> 8192 a', axes_lengths=(), name='w.unembed_period_sample', shape=(8192, 256), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), (0, 2))), out_axes=(0, 1), name='logits_period_sample', shape=(8192, 35), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), op_string='b0 a0 -> b0 a0', axes_lengths=(), name='logits_period_sample_rearrange_for_add_1', shape=(8192, 35), is_constant=True, is_explicitly_computable=True, can_be_sampled=True): -1.0}), name='logit_diff_all_seq_sample', shape=(8192, 35), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), index=(slice(None, None, None), -1), name='logit_diff_sample', hash_tensor_idx_by_value=False, shape=(8192,), is_constant=True, is_explicitly_computable=True, can_be_sampled=True)))(Einsum(args=((c3, (0,)), (c4, (0,))), out_axes=(), name='is_p_var_sample_sample_expectation', shape=(), is_constant=True, is_explicitly_computable=True, can_be_sampled=True)))(Index(node=ArrayConstant(value=torch.randn(torch.Size([48110])), shape=(48110,), uuid=UUID('dfe273fe-ae36-45fd-b32d-97eaa2556223'), name='is_p'), index=(torch.randint(1,48094,torch.Size([8192])),), name='is_p_var_sample', hash_tensor_idx_by_value=False, shape=(8192,), is_constant=True, is_explicitly_computable=True, can_be_sampled=True)))(Einsum(args=((c2, (0,)), (GeneralFunction(node=c1, function=reciprocal_fn, get_jacobian=reciprocal_jacobian, name='Ein_recip', allows_batching=True, non_batch_dims=(), shape=(), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), ())), out_axes=(0,), name='ScalarMul', shape=(8192,), is_constant=True, is_explicitly_computable=True, can_be_sampled=True)))(Add(items=FrozenDict({UnaryRearrange(node=ArrayConstant(value=torch.randn(torch.Size([256])), shape=(8192, 256), uuid=UUID('a3d94a03-5e58-4253-bef7-bb424a117c21'), name='final.n.w.bias_sample'), op_string='b0 a0 -> b0 o0 a0', axes_lengths=(('o0', 1),), name='final.n.w.bias_sample_rearrange_for_add_0', shape=(8192, 1, 256), is_constant=True, is_explicitly_computable=True, can_be_sampled=True): 1.0, UnaryRearrange(node=Einsum(args=((Einsum(args=((c9, (0, 1, 2)), (ArrayConstant(value=torch.randn(torch.Size([256])), shape=(8192, 256), uuid=UUID('ef0b3f2f-cdb5-414a-9f9e-e80eec300d86'), name='final.n.w.scale_sample'), (0, 2))), out_axes=(0, 1, 2), name='final.n.y_scale_sample', shape=(8192, 35, 256), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), (0, 1, 2)), (GeneralFunction(node=Add(items=FrozenDict({UnaryRearrange(node=Einsum(args=((c9, (0, 1, 2)), (c9, (0, 1, 2)), (FloatConstant(value=0.00390625, shape=(8192,), name='final.n.c.recip_h_size_sample'), (0,))), out_axes=(0, 1), name='final.n.var_sample', shape=(8192, 35), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), op_string='b0 a0 -> b0 a0', axes_lengths=(), name='final.n.var_sample_rearrange_for_add_0', shape=(8192, 35), is_constant=True, is_explicitly_computable=True, can_be_sampled=True): 1.0, UnaryRearrange(node=FloatConstant(value=1e-05, shape=(8192,), name='final.n.eps_sample'), op_string='b0 -> b0 o0', axes_lengths=(('o0', 1),), name='final.n.eps_sample_rearrange_for_add_1', shape=(8192, 1), is_constant=True, is_explicitly_computable=True, can_be_sampled=True): 1.0}), name='final.n.c.var_p_eps_sample', shape=(8192, 35), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), function=rsqrt_fn, get_jacobian=rsqrt_jacobian, name='final.n.full_mul_sample', allows_batching=True, non_batch_dims=(), shape=(8192, 35), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), (0, 1))), out_axes=(0, 1, 2), name='final.n.y_out_sample', shape=(8192, 35, 256), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), op_string='b0 a0 a1 -> b0 a0 a1', axes_lengths=(), name='final.n.y_out_sample_rearrange_for_add_1', shape=(8192, 35, 256), is_constant=True, is_explicitly_computable=True, can_be_sampled=True): 1.0}), name='final.n_sample', shape=(8192, 35, 256), is_constant=True, is_explicitly_computable=True, can_be_sampled=True)))(Einsum(args=((c2, (0,)),), out_axes=(), name='Ein', shape=(), is_constant=True, is_explicitly_computable=True, can_be_sampled=True)))(ArrayConstant(value=torch.randn(torch.Size([50259, 256])), shape=(50259, 256), uuid=UUID('af1caa2f-85d3-462f-8efe-f02dd1491924'), name='w.unembed')))(Einsum(args=((Index(node=AutoTag(node=FloatConstant(value=2.0785699438786117e-05, shape=(48110,), name='uniform_probs'), uuid=UUID('43012443-725a-4f51-aa91-2310a20c8db4'), name='uniform_probs_and_group'), index=(torch.randint(1,48094,torch.Size([8192])),), name='uniform_probs_and_group_importance_sample_idxed_probs', hash_tensor_idx_by_value=False, shape=(8192,), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), (0,)), (ArrayConstant(value=torch.randn(torch.Size([8192])), shape=(8192,), uuid=UUID('453a0a27-3c62-482a-b0b5-c5dde5099c67'), name='importance_sample_recip_sample_weights'), (0,))), out_axes=(0,), name='uniform_probs_and_group_importance_sample_weights', shape=(8192,), is_constant=True, is_explicitly_computable=True, can_be_sampled=True)))(Einsum(args=((Add(items=FrozenDict({c10: 1.0, c11: 1.0, UnaryRearrange(node=Einsum(args=((Einsum(args=((GeneralFunction(node=Add(items=FrozenDict({UnaryRearrange(node=Einsum(args=((Einsum(args=((Add(items=FrozenDict({UnaryRearrange(node=Einsum(args=((ArrayConstant(value=torch.randn(torch.Size([8, 32, 256])), shape=(8192, 8, 32, 256), uuid=UUID('04680ebd-ca99-4597-80b4-9fcdfb4f541d'), name='a1.w.q_sample'), (0, 1, 2, 3)), (c12, (0, 4, 3))), out_axes=(0, 1, 4, 2), name='a1.q_sample', shape=(8192, 8, 35, 32), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), op_string='b0 a0 a1 a2 -> b0 a0 a1 a2', axes_lengths=(), name='a1.q_sample_rearrange_for_add_0', shape=(8192, 8, 35, 32), is_constant=True, is_explicitly_computable=True, can_be_sampled=True): 1.0, UnaryRearrange(node=UnaryRearrange(node=Einsum(args=((ArrayConstant(value=torch.randn(torch.Size([8, 32, 256])), shape=(8, 32, 256), uuid=UUID('04680ebd-ca99-4597-80b4-9fcdfb4f541d'), name='a1.w.q'), (0, 1, 2)), (c14, (3, 2))), out_axes=(0, 3, 1), name='a1.w.pos_q', shape=(8, 35, 32), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), op_string='a b c -> 8192 a b c', axes_lengths=(), name='a1.w.pos_q_sample', shape=(8192, 8, 35, 32), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), op_string='b0 a0 a1 a2 -> b0 a0 a1 a2', axes_lengths=(), name='a1.w.pos_q_sample_rearrange_for_add_1', shape=(8192, 8, 35, 32), is_constant=True, is_explicitly_computable=True, can_be_sampled=True): 1.0}), name='a1.f_q_sample', shape=(8192, 8, 35, 32), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), (0, 1, 2, 3)), (Add(items=FrozenDict({UnaryRearrange(node=Einsum(args=((ArrayConstant(value=torch.randn(torch.Size([8, 32, 256])), shape=(8192, 8, 32, 256), uuid=UUID('4c311ff7-6006-4f1b-8df0-0df38a352e5d'), name='a1.w.k_sample'), (0, 1, 2, 3)), (c12, (0, 4, 3))), out_axes=(0, 1, 4, 2), name='a1.k_sample', shape=(8192, 8, 35, 32), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), op_string='b0 a0 a1 a2 -> b0 a0 a1 a2', axes_lengths=(), name='a1.k_sample_rearrange_for_add_0', shape=(8192, 8, 35, 32), is_constant=True, is_explicitly_computable=True, can_be_sampled=True): 1.0, UnaryRearrange(node=UnaryRearrange(node=Einsum(args=((ArrayConstant(value=torch.randn(torch.Size([8, 32, 256])), shape=(8, 32, 256), uuid=UUID('4c311ff7-6006-4f1b-8df0-0df38a352e5d'), name='a1.w.k'), (0, 1, 2)), (c14, (3, 2))), out_axes=(0, 3, 1), name='a1.w.pos_k', shape=(8, 35, 32), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), op_string='a b c -> 8192 a b c', axes_lengths=(), name='a1.w.pos_k_sample', shape=(8192, 8, 35, 32), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), op_string='b0 a0 a1 a2 -> b0 a0 a1 a2', axes_lengths=(), name='a1.w.pos_k_sample_rearrange_for_add_1', shape=(8192, 8, 35, 32), is_constant=True, is_explicitly_computable=True, can_be_sampled=True): 1.0}), name='a1.f_k_sample', shape=(8192, 8, 35, 32), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), (0, 1, 4, 3)), (FloatConstant(value=0.17677669529663687, shape=(8192,), name='a1.c.div_head_size_sample'), (0,))), out_axes=(0, 1, 2, 4), name='a1.scores_not_masked_sample', shape=(8192, 8, 35, 35), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), (0, 1, 2, 3)), (ArrayConstant(value=torch.randn(torch.Size([35, 35])), shape=(8192, 35, 35), uuid=UUID('8923e0b9-aec5-4117-909d-92e59f0da512'), name='a1.c.score_mask_sample'), (0, 2, 3))), out_axes=(0, 1, 2, 3), name='a1.scores_mul_mask_sample', shape=(8192, 8, 35, 35), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), op_string='b0 a0 a1 a2 -> b0 a0 a1 a2', axes_lengths=(), name='a1.scores_mul_mask_sample_rearrange_for_add_0', shape=(8192, 8, 35, 35), is_constant=True, is_explicitly_computable=True, can_be_sampled=True): 1.0, UnaryRearrange(node=ArrayConstant(value=torch.randn(torch.Size([35, 35])), shape=(8192, 35, 35), uuid=UUID('2a2d96ec-b792-4c3d-8942-f02804ae5da0'), name='a1.c.score_neg_inf_bias_sample'), op_string='b0 a0 a1 -> b0 o0 a0 a1', axes_lengths=(('o0', 1),), name='a1.c.score_neg_inf_bias_sample_rearrange_for_add_1', shape=(8192, 1, 35, 35), is_constant=True, is_explicitly_computable=True, can_be_sampled=True): 1.0}), name='a1.scores_sample', shape=(8192, 8, 35, 35), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), function=softmax_fn, get_jacobian=softmax_jacobian, name='a1.probs_sample', allows_batching=True, non_batch_dims=(-1,), shape=(8192, 8, 35, 35), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), (0, 1, 2, 3)), (Einsum(args=((ArrayConstant(value=torch.randn(torch.Size([8, 32, 256])), shape=(8192, 8, 32, 256), uuid=UUID('30181721-36a6-4d46-b87b-4e87aea6aeef'), name='a1.w.v_sample'), (0, 1, 2, 3)), (c12, (0, 4, 3))), out_axes=(0, 1, 4, 2), name='a1.v_sample', shape=(8192, 8, 35, 32), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), (0, 1, 3, 4))), out_axes=(0, 2, 1, 4), name='a1.comb_v_sample', shape=(8192, 35, 8, 32), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), (0, 1, 2, 3)), (ArrayConstant(value=torch.randn(torch.Size([8, 32, 256])), shape=(8192, 8, 32, 256), uuid=UUID('67abd9d5-cb28-488e-916d-6e6aea3b4ca8'), name='a1.w.out_sample'), (0, 2, 3, 4))), out_axes=(0, 1, 4), name='a1_sample', shape=(8192, 35, 256), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), op_string='b0 a0 a1 -> b0 a0 a1', axes_lengths=(), name='a1_sample_rearrange_for_add_2', shape=(8192, 35, 256), is_constant=True, is_explicitly_computable=True, can_be_sampled=True): 1.0}), name='final.inp_sample', shape=(8192, 35, 256), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), (0, 1, 2)), (ArrayConstant(value=torch.randn(torch.Size([256, 256])), shape=(8192, 256, 256), uuid=UUID('9bd04993-d6d5-46c8-8265-6d6db412d94b'), name='final.n.c.sub_mean_sample'), (0, 2, 3))), out_axes=(0, 1, 3), name='final.n.z_mean_sample', shape=(8192, 35, 256), is_constant=True, is_explicitly_computable=True, can_be_sampled=True)))(Add(items=FrozenDict({UnaryRearrange(node=ArrayConstant(value=torch.randn(torch.Size([256])), shape=(8192, 256), uuid=UUID('08cfad93-cdc4-446a-bd10-ffa8e2ec7177'), name='a1.n.w.bias_sample'), op_string='b0 a0 -> b0 o0 a0', axes_lengths=(('o0', 1),), name='a1.n.w.bias_sample_rearrange_for_add_0', shape=(8192, 1, 256), is_constant=True, is_explicitly_computable=True, can_be_sampled=True): 1.0, UnaryRearrange(node=Einsum(args=((Einsum(args=((c13, (0, 1, 2)), (ArrayConstant(value=torch.randn(torch.Size([256])), shape=(8192, 256), uuid=UUID('b39cc991-01a2-4355-b6f1-ffe85444ee6a'), name='a1.n.w.scale_sample'), (0, 2))), out_axes=(0, 1, 2), name='a1.n.y_scale_sample', shape=(8192, 35, 256), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), (0, 1, 2)), (GeneralFunction(node=Add(items=FrozenDict({UnaryRearrange(node=Einsum(args=((c13, (0, 1, 2)), (c13, (0, 1, 2)), (FloatConstant(value=0.00390625, shape=(8192,), name='a1.n.c.recip_h_size_sample'), (0,))), out_axes=(0, 1), name='a1.n.var_sample', shape=(8192, 35), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), op_string='b0 a0 -> b0 a0', axes_lengths=(), name='a1.n.var_sample_rearrange_for_add_0', shape=(8192, 35), is_constant=True, is_explicitly_computable=True, can_be_sampled=True): 1.0, UnaryRearrange(node=FloatConstant(value=1e-05, shape=(8192,), name='a1.n.eps_sample'), op_string='b0 -> b0 o0', axes_lengths=(('o0', 1),), name='a1.n.eps_sample_rearrange_for_add_1', shape=(8192, 1), is_constant=True, is_explicitly_computable=True, can_be_sampled=True): 1.0}), name='a1.n.c.var_p_eps_sample', shape=(8192, 35), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), function=rsqrt_fn, get_jacobian=rsqrt_jacobian, name='a1.n.full_mul_sample', allows_batching=True, non_batch_dims=(), shape=(8192, 35), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), (0, 1))), out_axes=(0, 1, 2), name='a1.n.y_out_sample', shape=(8192, 35, 256), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), op_string='b0 a0 a1 -> b0 a0 a1', axes_lengths=(), name='a1.n.y_out_sample_rearrange_for_add_1', shape=(8192, 35, 256), is_constant=True, is_explicitly_computable=True, can_be_sampled=True): 1.0}), name='a1.n_sample', shape=(8192, 35, 256), is_constant=True, is_explicitly_computable=True, can_be_sampled=True)))(Einsum(args=((Add(items=FrozenDict({c10: 1.0, c11: 1.0}), name='a1.inp_sample', shape=(8192, 35, 256), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), (0, 1, 2)), (ArrayConstant(value=torch.randn(torch.Size([256, 256])), shape=(8192, 256, 256), uuid=UUID('15dad660-1d78-4801-96de-9bfed51d4948'), name='a1.n.c.sub_mean_sample'), (0, 2, 3))), out_axes=(0, 1, 3), name='a1.n.z_mean_sample', shape=(8192, 35, 256), is_constant=True, is_explicitly_computable=True, can_be_sampled=True)))(UnaryRearrange(node=c17, op_string='b0 a0 a1 -> b0 a0 a1', axes_lengths=(), name='embed_var_sample_rearrange_for_add_0', shape=(8192, 35, 256), is_constant=True, is_explicitly_computable=True, can_be_sampled=True)))(UnaryRearrange(node=Einsum(args=((Einsum(args=((GeneralFunction(node=Add(items=FrozenDict({UnaryRearrange(node=Einsum(args=((Einsum(args=((Add(items=FrozenDict({UnaryRearrange(node=Einsum(args=((ArrayConstant(value=torch.randn(torch.Size([8, 32, 256])), shape=(8192, 8, 32, 256), uuid=UUID('a3982e07-149f-4c74-a4e5-20bc7f599ef9'), name='a0.w.q_sample'), (0, 1, 2, 3)), (c15, (0, 4, 3))), out_axes=(0, 1, 4, 2), name='a0.q_sample', shape=(8192, 8, 35, 32), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), op_string='b0 a0 a1 a2 -> b0 a0 a1 a2', axes_lengths=(), name='a0.q_sample_rearrange_for_add_0', shape=(8192, 8, 35, 32), is_constant=True, is_explicitly_computable=True, can_be_sampled=True): 1.0, UnaryRearrange(node=UnaryRearrange(node=Einsum(args=((ArrayConstant(value=torch.randn(torch.Size([8, 32, 256])), shape=(8, 32, 256), uuid=UUID('a3982e07-149f-4c74-a4e5-20bc7f599ef9'), name='a0.w.q'), (0, 1, 2)), (c14, (3, 2))), out_axes=(0, 3, 1), name='a0.w.pos_q', shape=(8, 35, 32), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), op_string='a b c -> 8192 a b c', axes_lengths=(), name='a0.w.pos_q_sample', shape=(8192, 8, 35, 32), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), op_string='b0 a0 a1 a2 -> b0 a0 a1 a2', axes_lengths=(), name='a0.w.pos_q_sample_rearrange_for_add_1', shape=(8192, 8, 35, 32), is_constant=True, is_explicitly_computable=True, can_be_sampled=True): 1.0}), name='a0.f_q_sample', shape=(8192, 8, 35, 32), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), (0, 1, 2, 3)), (Add(items=FrozenDict({UnaryRearrange(node=Einsum(args=((ArrayConstant(value=torch.randn(torch.Size([8, 32, 256])), shape=(8192, 8, 32, 256), uuid=UUID('f03fd2a7-a276-4c9c-afd5-d0d0914861b8'), name='a0.w.k_sample'), (0, 1, 2, 3)), (c15, (0, 4, 3))), out_axes=(0, 1, 4, 2), name='a0.k_sample', shape=(8192, 8, 35, 32), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), op_string='b0 a0 a1 a2 -> b0 a0 a1 a2', axes_lengths=(), name='a0.k_sample_rearrange_for_add_0', shape=(8192, 8, 35, 32), is_constant=True, is_explicitly_computable=True, can_be_sampled=True): 1.0, UnaryRearrange(node=UnaryRearrange(node=Einsum(args=((ArrayConstant(value=torch.randn(torch.Size([8, 32, 256])), shape=(8, 32, 256), uuid=UUID('f03fd2a7-a276-4c9c-afd5-d0d0914861b8'), name='a0.w.k'), (0, 1, 2)), (c14, (3, 2))), out_axes=(0, 3, 1), name='a0.w.pos_k', shape=(8, 35, 32), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), op_string='a b c -> 8192 a b c', axes_lengths=(), name='a0.w.pos_k_sample', shape=(8192, 8, 35, 32), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), op_string='b0 a0 a1 a2 -> b0 a0 a1 a2', axes_lengths=(), name='a0.w.pos_k_sample_rearrange_for_add_1', shape=(8192, 8, 35, 32), is_constant=True, is_explicitly_computable=True, can_be_sampled=True): 1.0}), name='a0.f_k_sample', shape=(8192, 8, 35, 32), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), (0, 1, 4, 3)), (FloatConstant(value=0.17677669529663687, shape=(8192,), name='a0.c.div_head_size_sample'), (0,))), out_axes=(0, 1, 2, 4), name='a0.scores_not_masked_sample', shape=(8192, 8, 35, 35), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), (0, 1, 2, 3)), (ArrayConstant(value=torch.randn(torch.Size([35, 35])), shape=(8192, 35, 35), uuid=UUID('251501a5-b371-42e4-8853-887bba31173c'), name='a0.c.score_mask_sample'), (0, 2, 3))), out_axes=(0, 1, 2, 3), name='a0.scores_mul_mask_sample', shape=(8192, 8, 35, 35), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), op_string='b0 a0 a1 a2 -> b0 a0 a1 a2', axes_lengths=(), name='a0.scores_mul_mask_sample_rearrange_for_add_0', shape=(8192, 8, 35, 35), is_constant=True, is_explicitly_computable=True, can_be_sampled=True): 1.0, UnaryRearrange(node=ArrayConstant(value=torch.randn(torch.Size([35, 35])), shape=(8192, 35, 35), uuid=UUID('09bec2c1-91bd-48c1-a940-3536620bfad8'), name='a0.c.score_neg_inf_bias_sample'), op_string='b0 a0 a1 -> b0 o0 a0 a1', axes_lengths=(('o0', 1),), name='a0.c.score_neg_inf_bias_sample_rearrange_for_add_1', shape=(8192, 1, 35, 35), is_constant=True, is_explicitly_computable=True, can_be_sampled=True): 1.0}), name='a0.scores_sample', shape=(8192, 8, 35, 35), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), function=softmax_fn, get_jacobian=softmax_jacobian, name='a0.probs_sample', allows_batching=True, non_batch_dims=(-1,), shape=(8192, 8, 35, 35), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), (0, 1, 2, 3)), (Einsum(args=((ArrayConstant(value=torch.randn(torch.Size([8, 32, 256])), shape=(8192, 8, 32, 256), uuid=UUID('1b4daf99-41df-43e3-a390-25abda812c6e'), name='a0.w.v_sample'), (0, 1, 2, 3)), (c15, (0, 4, 3))), out_axes=(0, 1, 4, 2), name='a0.v_sample', shape=(8192, 8, 35, 32), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), (0, 1, 3, 4))), out_axes=(0, 2, 1, 4), name='a0.comb_v_sample', shape=(8192, 35, 8, 32), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), (0, 1, 2, 3)), (ArrayConstant(value=torch.randn(torch.Size([8, 32, 256])), shape=(8192, 8, 32, 256), uuid=UUID('80972fcd-7330-47d2-a0a4-b6ce8b2b3f0f'), name='a0.w.out_sample'), (0, 2, 3, 4))), out_axes=(0, 1, 4), name='a0_sample', shape=(8192, 35, 256), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), op_string='b0 a0 a1 -> b0 a0 a1', axes_lengths=(), name='a0_sample_rearrange_for_add_1', shape=(8192, 35, 256), is_constant=True, is_explicitly_computable=True, can_be_sampled=True)))(Add(items=FrozenDict({UnaryRearrange(node=ArrayConstant(value=torch.randn(torch.Size([256])), shape=(8192, 256), uuid=UUID('c9c9b86e-0121-4186-bc51-90bc4309bdd3'), name='a0.n.w.bias_sample'), op_string='b0 a0 -> b0 o0 a0', axes_lengths=(('o0', 1),), name='a0.n.w.bias_sample_rearrange_for_add_0', shape=(8192, 1, 256), is_constant=True, is_explicitly_computable=True, can_be_sampled=True): 1.0, UnaryRearrange(node=Einsum(args=((Einsum(args=((c16, (0, 1, 2)), (ArrayConstant(value=torch.randn(torch.Size([256])), shape=(8192, 256), uuid=UUID('340f89bc-c0de-4869-abc1-b30ab43dc551'), name='a0.n.w.scale_sample'), (0, 2))), out_axes=(0, 1, 2), name='a0.n.y_scale_sample', shape=(8192, 35, 256), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), (0, 1, 2)), (GeneralFunction(node=Add(items=FrozenDict({UnaryRearrange(node=Einsum(args=((c16, (0, 1, 2)), (c16, (0, 1, 2)), (FloatConstant(value=0.00390625, shape=(8192,), name='a0.n.c.recip_h_size_sample'), (0,))), out_axes=(0, 1), name='a0.n.var_sample', shape=(8192, 35), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), op_string='b0 a0 -> b0 a0', axes_lengths=(), name='a0.n.var_sample_rearrange_for_add_0', shape=(8192, 35), is_constant=True, is_explicitly_computable=True, can_be_sampled=True): 1.0, UnaryRearrange(node=FloatConstant(value=1e-05, shape=(8192,), name='a0.n.eps_sample'), op_string='b0 -> b0 o0', axes_lengths=(('o0', 1),), name='a0.n.eps_sample_rearrange_for_add_1', shape=(8192, 1), is_constant=True, is_explicitly_computable=True, can_be_sampled=True): 1.0}), name='a0.n.c.var_p_eps_sample', shape=(8192, 35), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), function=rsqrt_fn, get_jacobian=rsqrt_jacobian, name='a0.n.full_mul_sample', allows_batching=True, non_batch_dims=(), shape=(8192, 35), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), (0, 1))), out_axes=(0, 1, 2), name='a0.n.y_out_sample', shape=(8192, 35, 256), is_constant=True, is_explicitly_computable=True, can_be_sampled=True), op_string='b0 a0 a1 -> b0 a0 a1', axes_lengths=(), name='a0.n.y_out_sample_rearrange_for_add_1', shape=(8192, 35, 256), is_constant=True, is_explicitly_computable=True, can_be_sampled=True): 1.0}), name='a0.n_sample', shape=(8192, 35, 256), is_constant=True, is_explicitly_computable=True, can_be_sampled=True)))(ArrayConstant(value=torch.randn(torch.Size([35, 256])), shape=(35, 256), uuid=UUID('4382a439-ba87-43ff-8dd7-13f9f3d0c147'), name='w.pos_embeds')))(Einsum(args=((c17, (0, 1, 2)), (ArrayConstant(value=torch.randn(torch.Size([256, 256])), shape=(8192, 256, 256), uuid=UUID('87fea373-a043-4a30-a0b3-15d096b6d0db'), name='a0.n.c.sub_mean_sample'), (0, 2, 3))), out_axes=(0, 1, 3), name='a0.n.z_mean_sample', shape=(8192, 35, 256), is_constant=True, is_explicitly_computable=True, can_be_sampled=True)))(Index(node=ArrayConstant(value=torch.randn(torch.Size([48110, 35, 256])), shape=(48110, 35, 256), uuid=UUID('05a4b3da-31f1-4610-a3b9-58f483c9bac2'), name='unnamed'), index=(torch.randint(1,48094,torch.Size([8192])),), name='embed_var_sample', hash_tensor_idx_by_value=False, shape=(8192, 35, 256), is_constant=True, is_explicitly_computable=True, can_be_sampled=True))
"""
)
dtype = torch.float32
nb_example = cast_circuit(nb_example, device="cuda", dtype=dtype)
nb_example_rust = deep_canonicalize(strip_names(py_to_rust(nb_example)))
print(f"py circ has {count_nodes(nb_example)} nodes")
nb_example_rust.print_stats()
nb_example_rust.compiler_print()
with catchtime("py"):
    py_simp = optimize_with_backreferences(circuit_to_canonical_ast(nb_example))[0]
with catchtime("simp"):
    simped = optimize_circuit(nb_example_rust, OptimizationSettings())
simped.print_stats()
simped.compiler_print()
print(f"py circ has {ast_count_nodes(py_simp)} nodes, rust circ has {r_count_nodes(simped)} nodes")
with catchtime("scheduled eval"):
    sched_eval = optimize_and_evaluate(simped, OptimizationSettings(0, 1_800_000_000, scheduling_simplify=False))
with catchtime("batch eval"):
    batched = deep_heuristic_nest_adds(deep_optimize_einsums(batch_largest(simped, 4, 50_000_000)))
    batched.compiler_print()
    print("batched num nodes", r_count_nodes(batched))
    sched_eval = scheduled_evaluate(batched, OptimizationSettings(2, 2_000_000_000, scheduling_simplify=False))
with catchtime("rust circuit eval"):
    rust_eval = simped.evaluate()
with catchtime("rust raw eval"):
    rust_eval_raw = nb_example_rust.evaluate()
python_eval = evaluate_circuit(nb_example, verbose=0, device="cuda", dtype=dtype)
with catchtime("memf"):
    MemoizedFn(evaluate_fn(device="cuda", dtype=dtype))(nb_example)
torch.testing.assert_close(python_eval, rust_eval_raw)
deep_canonicalize(compiler_simp(nb_example_rust)).compiler_print()
print_ast_deep(py_simp)

# current rust flops: 25T python flops: 700B



# this file is included with `include_str!()` in py_types.rs
import functools
import itertools
from uuid import UUID
import torch
import torch.nn.functional
from typing import List, Tuple
import math

none = None


def check_type(x, type_v):
    if not isinstance(x, type_v):
        raise TypeError(f"expected type {repr(type_v)} but got {type(x)}")


def get_uuid_bytes(x):
    check_type(x, UUID)
    return x.bytes


def construct_uuid_bytes(b):
    return UUID(bytes=b)


def get_tensor_shape(x):
    check_type(x, torch.Tensor)
    return list(x.size())


def get_id(x):
    return id(x)


def cast_int(x):
    return int(x)


make_bytes = bytes

dtype_from_string = {
    "float32": torch.float32,
    "float64": torch.float64,
    "float16": torch.float16,
    "int64": torch.int64,
    "int32": torch.int32,
    "int16": torch.int16,
    "int8": torch.int8,
}


def scalar_to_tensor(scalar, shape, device_dtype):
    "Makes a stride 0 repeat view into a 1 element tensor"
    scalar_tensor = torch.tensor(scalar, device=device_dtype.device, dtype=dtype_from_string[device_dtype.dtype])
    return torch.broadcast_to(scalar_tensor, tuple(shape))


def tensor_scale(tensor):
    return tensor.abs().mean().cpu().item()


def cast_tensor(tensor, device_dtype):
    return tensor.to(device=device_dtype.device, dtype=dtype_from_string[device_dtype.dtype])


zero_tensor = torch.zeros(())


def un_flat_concat(tensor: torch.Tensor, shapes) -> List[torch.Tensor]:
    lens = [math.prod(x) for x in shapes]
    flats = torch.split(tensor, lens, dim=0)
    return [x.reshape(shape) for x, shape in zip(flats, shapes)]


def log_exp_p_1_fn(x: torch.Tensor):
    # piecewise to nicely handle numerics
    addr = 1.0
    return torch.where(x < 0.0, torch.log(torch.exp(x) + addr), torch.log(1.0 + torch.exp(-x) * addr) + x)


generalfunctions = {
    "gelu": torch.nn.functional.gelu,
    "relu": torch.relu,
    "step": lambda x: torch.where(
        x > 0.0, torch.ones((), dtype=x.dtype, device=x.device), torch.zeros((), dtype=x.dtype, device=x.device)
    ),
    "sigmoid": torch.sigmoid,
    "rsqrt": torch.rsqrt,
    "reciprocal": torch.reciprocal,
    "tanh": torch.tanh,
    "softmax": functools.partial(torch.softmax, dim=-1),
    "log_softmax": functools.partial(torch.log_softmax, dim=-1),
    "log_exp_p_1": log_exp_p_1_fn,
    "gaussian_pdf": lambda x: torch.exp(-(x ** 2) / 2) / math.sqrt(2 * math.pi),
    "gaussian_cdf": lambda x: torch.distributions.normal.Normal(0, 1).cdf(x),
    "q_from_qr": lambda x: torch.linalg.qr(x)[0],
}


def make_diagonal(tensor: torch.Tensor, tensor_ints: Tuple[int, ...], out_ints: Tuple[int, ...]):
    int_sizes = {x: tensor.shape[i] for i, x in enumerate(tensor_ints)}
    deduped_shape = [int_sizes[x] for x in tensor_ints]
    result = torch.zeros(tuple(int_sizes[x] for x in out_ints), dtype=tensor.dtype, device=tensor.device)
    normal_strides_out = result.stride()
    fancy_strides = []
    for i in tensor_ints:
        indices = [j for j, x in enumerate(out_ints) if x == i]
        strides = [normal_strides_out[k] for k in indices]
        stride_here = sum(strides)
        fancy_strides.append(stride_here)
    fancy_strided = torch.as_strided(result, deduped_shape, fancy_strides)
    fancy_strided += tensor
    return result


einsum = lambda tensors_and_axes, out_axes: torch.einsum(*itertools.chain(*tensors_and_axes), out_axes)



from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from uuid import UUID

import torch

Shape = Tuple[int, ...]

class EinsumSpec(object):
    input_ints: List[List[int]]
    output_ints: List[int]
    int_sizes: List[int]
    def __init__(
        self,
        input_ints: List[List[int]],
        output_ints: List[int],
        int_sizes: Dict[int, int],
    ): ...
    def flops(self) -> int: ...
    def optimize_dp(
        self,
        check_outer: bool,
        mem_limit: Optional[int] = None,
        hash_limit: Optional[int] = None,
    ) -> List[List[int]]: ...
    def optimize(
        self,
        check_outer: bool,
        mem_limit: Optional[int] = None,
        hash_limit: Optional[int] = None,
    ) -> List[List[int]]: ...
    def normalize(self) -> "EinsumSpec": ...
    def validate(self) -> bool: ...
    def shapes(self) -> Tuple[List[List[int]], List[int]]: ...
    def to_einsum_string(self) -> str: ...
    @staticmethod
    def string_to_ints(string: str) -> Tuple[List[List[int]], List[int]]: ...
    def symbol_circuit(self) -> Einsum: ...

def optimize_einsum_spec_cached(
    spec: EinsumSpec,
    check_outer: bool,
    mem_limit: Optional[int] = None,
    hash_limit: Optional[int] = None,
) -> List[List[int]]: ...

class RearrangeSpec(object):
    input_ints: List[List[int]]
    output_ints: List[List[int]]
    int_sizes: List[Optional[int]]
    def __init__(
        self,
        input_ints: List[List[int]],
        output_ints: List[List[int]],
        int_sizes: List[Optional[int]],
    ): ...
    def is_identity(self) -> bool: ...
    def is_permute(self) -> bool: ...
    def shapes(self) -> Tuple[List[int], List[int]]: ...
    def is_valid(self) -> bool: ...
    def to_einops_string(self) -> str: ...
    def to_einops_string_and_letter_sizes(
        self,
    ) -> Tuple[str, List[Tuple[str, int]]]: ...
    def apply(self, tensor: torch.Tensor) -> torch.Tensor: ...
    @staticmethod
    def fuse(inner: RearrangeSpec, outer: RearrangeSpec) -> RearrangeSpec: ...
    def canonicalize(self, special_case_ones: bool = True) -> RearrangeSpec: ...
    def fill_empty_ints(self, allow_rust_invalid: bool) -> RearrangeSpec: ...
    def conform_to_input_shape(self, shape: Tuple[int, ...], coerce: bool) -> RearrangeSpec: ...
    @staticmethod
    def ident(rank: int) -> RearrangeSpec: ...
    def to_py_rearrange_spec(self, shape: Tuple[int, ...]) -> Any: ...
    @staticmethod
    def flatten(rank: int) -> RearrangeSpec: ...
    @staticmethod
    def unflatten(shape: List[int]) -> RearrangeSpec: ...

class Circuit:
    @property
    def shape(self) -> Shape: ...
    @property
    def is_constant(self) -> bool: ...
    @property
    def is_explicitly_computable(self) -> bool: ...
    @property
    def can_be_sampled(self) -> bool: ...
    @property
    def name(self) -> str: ...
    @property
    def hash(self) -> bytes: ...
    def children(self) -> List[Circuit]: ...
    def self_flops(self) -> int: ...
    def total_flops(self) -> int: ...
    def max_non_input_size(self) -> int: ...
    def apply_fn_to_sub(self, f: Callable[[Circuit], Circuit]) -> Circuit: ...
    def compiler_print(self) -> None: ...
    def print_stats(self) -> None: ...
    def compiler_repr(self) -> str: ...
    def numel(self) -> int: ...
    def rank(self) -> int: ...
    def to_py(self) -> Any: ...
    def evaluate(self, device_dtype: TorchDeviceDtypeOp = TorchDeviceDtypeOp()) -> torch.Tensor: ...

class ArrayConstant(Circuit):
    def __init__(self, value: torch.Tensor, name: Optional[str] = "ArrayConstant") -> None: ...
    @property
    def uuid(self) -> UUID: ...
    @property
    def value(self) -> torch.Tensor: ...
    @staticmethod
    def randn(*args: int) -> "ArrayConstant": ...
    @staticmethod
    def randn_named(shape: List[int], name: Optional[str], device_dtype: TorchDeviceDtypeOp) -> "ArrayConstant": ...
    @staticmethod
    def randn_seeded(
        shape: List[int], name: Optional[str], device_dtype: TorchDeviceDtypeOp, seed: int
    ) -> "ArrayConstant": ...
    @staticmethod
    def new_named_axes(
        value: torch.Tensor, name: Optional[str], named_axes: List[Optional[str]]
    ) -> "ArrayConstant": ...

class Symbol(Circuit):
    def __init__(self, shape: Shape, uuid: UUID, name: Optional[str] = "Symbol") -> None: ...
    @property
    def uuid(self) -> UUID: ...

class ScalarConstant(Circuit):
    def __init__(self, value: float, shape: Shape = (), name: Optional[str] = "ScalarConstant") -> None: ...
    @property
    def value(self) -> float: ...
    def is_zero(self) -> bool: ...
    def is_one(self) -> bool: ...

class Einsum(Circuit):
    def __init__(
        self,
        *args: Tuple[Circuit, Tuple[int, ...]],
        out_axes: Tuple[int, ...],
        name: Optional[str] = None,
    ) -> None: ...
    @property
    def args(self) -> List[Tuple[Circuit, Tuple[int, ...]]]: ...
    @property
    def out_axes(self) -> Tuple[int, ...]: ...
    def all_input_circuits(self) -> List[Circuit]: ...
    def all_input_axes(self) -> List[Tuple[int, ...]]: ...
    @staticmethod
    def from_einsum_string(string: str, nodes: List[Circuit], name: Optional[str] = None) -> Einsum: ...
    @staticmethod
    def from_spec(spec: EinsumSpec, circuits: List[Circuit], name: Optional[str] = None) -> Einsum: ...
    @staticmethod
    def new_diag(node: Circuit, ints: List[int], name: Optional[str] = None) -> Einsum: ...
    @staticmethod
    def new_trace(node: Circuit, ints: List[int], name: Optional[str] = None) -> Einsum: ...

class Add(Circuit):
    def __init__(self, nodes: List[Circuit], name: Optional[str] = None) -> None: ...
    @property
    def nodes(self) -> List[Circuit]: ...
    def has_broadcast(self) -> bool: ...
    def nodes_and_rank_differences(self) -> List[Tuple[Circuit, int]]: ...

class Rearrange(Circuit):
    def __init__(self, node: Circuit, spec: RearrangeSpec, name: Optional[str] = None) -> None: ...
    @property
    def node(self) -> Circuit: ...
    @property
    def spec(self) -> RearrangeSpec: ...

class Index(Circuit):
    def __init__(
        self,
        node: Circuit,
        index: Tuple[Union[int, slice, torch.Tensor], ...],
        name: Optional[str] = None,
    ) -> None: ...
    @property
    def node(self) -> Circuit: ...
    @property
    def index(self) -> Tuple[Union[int, slice, torch.Tensor], ...]: ...

class Scatter(Circuit):
    def __init__(
        self,
        node: Circuit,
        index: Tuple[Union[int, slice, torch.Tensor], ...],
        shape: Tuple[int, ...],
        name: Optional[str] = None,
    ) -> None: ...
    @property
    def node(self) -> Circuit: ...
    @property
    def index(self) -> Tuple[Union[int, slice, torch.Tensor], ...]: ...
    def is_identity(self) -> bool: ...

class GeneralFunctionSpec(object):
    function: Callable
    get_shape: Callable
    get_jacobian: Optional[Callable]
    num_non_batchable_output_dims: int
    input_batchability: List[bool]
    name: str
    def __init__(
        self,
        function: Callable,
        get_shape: Callable,
        get_jacobian: Optional[Callable],
        num_non_batchable_output_dims: int,
        input_batchability: List[bool],
        name: str,
    ) -> None: ...
    def is_batchable(self) -> bool: ...

class PyGFSpecShapeGetter(object):
    num_non_batchable: int
    def __init__(self, num_non_batchable: int) -> None: ...
    def __call__(self, args: List[Shape]) -> Optional[Shape]: ...

class GeneralFunction(Circuit):
    def __init__(
        self,
        nodes: List[Circuit],
        spec: GeneralFunctionSpec,
        name: Optional[str] = None,
    ) -> None: ...
    @property
    def nodes(self) -> List[Circuit]: ...
    @property
    def spec(self) -> GeneralFunctionSpec: ...
    @staticmethod
    def new_by_name(nodes: List[Circuit], spec_name: str, name: Optional[str] = None) -> GeneralFunction: ...

class Concat(Circuit):
    def __init__(self, nodes: List[Circuit], axis: int, name: Optional[str] = None) -> None: ...
    @property
    def nodes(self) -> List[Circuit]: ...
    @property
    def axis(self) -> int: ...

class TorchDeviceDtype(object):
    def __init__(self, device: str, dtype: str) -> None: ...
    @property
    def device(self) -> str: ...
    @property
    def dtype(self) -> str: ...
    @staticmethod
    def from_tensor(tensor: torch.Tensor) -> TorchDeviceDtype: ...

class TorchDeviceDtypeOp(object):
    def __init__(self, device: Optional[str] = None, dtype: Optional[str] = None) -> None: ...
    @staticmethod
    def default() -> TorchDeviceDtypeOp: ...
    @property
    def device(self) -> Optional[str]: ...
    @property
    def dtype(self) -> Optional[str]: ...

class OptimizationSettings(object):
    verbose: int
    max_memory: int
    max_single_tensor_memory: int
    scheduling_num_mem_chunks: int
    distribute_min_size: Optional[int]
    scheduling_naive: bool
    scheduling_simplify: bool
    adjust_numerical_scale: bool
    numerical_scale_min: float
    numerical_scale_max: float
    capture_and_print: bool
    create_tests: bool
    log_simplifications: bool
    log_slow_einsums: bool
    def __init__(
        self,
        verbose=0,
        max_memory=9_000_000_000,
        max_single_tensor_memory=9_000_000_000,
        scheduling_num_mem_chunks=200,
        distribute_min_size=None,
        scheduling_naive=False,
        scheduling_simplify=True,
        adjust_numerical_scale=False,
        numerical_scale_min=1e-8,
        numerical_scale_max=1e8,
        capture_and_print=False,
        create_tests=False,
        log_simplifications=False,
        log_slow_einsums=False,
    ) -> None: ...

class PyOOMError(Exception): ...

class Schedule(object):
    @property
    def device_dtype(self) -> TorchDeviceDtype: ...
    @property
    def constants(self) -> Dict[bytes, torch.Tensor]: ...
    @property
    def instructions(self) -> List[Union[bytes, Circuit]]: ...
    def replace_tensors(self, map: Dict[bytes, torch.Tensor]) -> Schedule: ...
    def map_tensors(self, f: Callable[[bytes, torch.Tensor], torch.Tensor]) -> Schedule: ...
    def get_stats(self) -> ScheduleStats: ...
    def evaluate(self, settings: OptimizationSettings) -> torch.Tensor: ...
    def evaluate_many(self, settings: OptimizationSettings) -> List[torch.Tensor]: ...

class ScheduleStats(object):
    @property
    def max_mem(self) -> int: ...
    @property
    def constant_mem(self) -> int: ...
    @property
    def max_circuit_set(self) -> Set[Circuit]: ...

def add_collapse_scalar_inputs(add: Add) -> Optional[Add]: ...
def add_deduplicate(add: Add) -> Optional[Add]: ...
def add_flatten_once(add: Add) -> Optional[Add]: ...
def remove_add_few_input(add: Add) -> Optional[Add]: ...
def add_pull_removable_axes(add: Add, remove_non_common_axes: bool) -> Optional[Circuit]: ...
def einsum_flatten_once(einsum: Einsum) -> Optional[Einsum]: ...
def einsum_elim_identity(einsum: Einsum) -> Optional[Circuit]: ...
def index_merge_scalar(index: Index) -> Optional[Circuit]: ...
def index_elim_identity(index: Index) -> Optional[Circuit]: ...
def index_fuse(index: Index) -> Optional[Index]: ...
def rearrange_fuse(node: Rearrange) -> Optional[Rearrange]: ...
def rearrange_merge_scalar(rearrange: Rearrange) -> Optional[Circuit]: ...
def rearrange_elim_identity(rearrange: Rearrange) -> Optional[Circuit]: ...
def concat_elim_identity(concat: Concat) -> Optional[Circuit]: ...
def concat_merge_uniform(concat: Concat) -> Optional[Concat]: ...
def concat_pull_removable_axes(concat: Concat) -> Optional[Circuit]: ...
def generalfunction_pull_removable_axes(node: GeneralFunction) -> Optional[Circuit]: ...
def einsum_pull_removable_axes(node: Einsum) -> Optional[Circuit]: ...
def scatter_pull_removable_axes(node: Scatter) -> Optional[Circuit]: ...
def add_make_broadcasts_explicit(node: Add) -> Optional[Add]: ...
def distribute(node: Einsum, operand_idx: int, do_broadcasts: bool) -> Optional[Add]: ...
def distribute_all(node: Einsum) -> Optional[Add]: ...
def einsum_of_permute_merge(node: Einsum) -> Optional[Einsum]: ...
def permute_of_einsum_merge(node: Rearrange) -> Optional[Einsum]: ...
def einsum_elim_zero(node: Einsum) -> Optional[ScalarConstant]: ...
def einsum_merge_scalars(node: Einsum) -> Optional[Einsum]: ...
def push_down_index(node: Index) -> Optional[Circuit]: ...
def deep_push_down_index(node: Circuit, min_size: Optional[int] = None) -> Circuit: ...
def index_split_axes(node: Index, top_axes: Set[int]) -> Optional[Index]: ...
def add_elim_zeros(node: Add) -> Optional[Add]: ...
def compiler_simp(node: Circuit) -> Circuit: ...
def compiler_simp_step(node: Circuit) -> Optional[Circuit]: ...
def compiler_simp_until_same(node: Circuit) -> Circuit: ...
def strip_names(node: Circuit) -> Circuit: ...
def deep_canonicalize(node: Circuit) -> Circuit: ...
def canonicalize_node(node: Circuit) -> Circuit: ...
def deep_maybe_distribute(node: Circuit, settings: OptimizationSettings) -> Circuit: ...
def maybe_distribute(node: Einsum, settings: OptimizationSettings) -> Optional[Circuit]: ...
def einsum_nest_optimize(node: Einsum) -> Optional[Einsum]: ...
def deep_optimize_einsums(node: Circuit) -> Circuit: ...
def einsum_nest_path(node: Einsum, path: List[List[int]]) -> Einsum: ...
def scatter_elim_identity(node: Scatter) -> Optional[Circuit]: ...
def einsum_pull_scatter(node: Einsum) -> Optional[Circuit]: ...
def add_pull_scatter(node: Add) -> Optional[Circuit]: ...
def index_einsum_to_scatter(node: Index) -> Optional[Circuit]: ...
def optimize_circuit(circuit: Circuit, settings: OptimizationSettings) -> Circuit: ...
def cast_circuit(circuit: Circuit, device_dtype: TorchDeviceDtype) -> Circuit: ...
def scatter_to_concat(scatter: Scatter) -> Circuit: ...
def count_nodes(circuit: Circuit) -> int: ...
def scheduled_evaluate(circuit: Circuit, settings: OptimizationSettings) -> torch.Tensor: ...
def optimize_and_evaluate(circuit: Circuit, settings: OptimizationSettings) -> torch.Tensor: ...
def optimize_and_evaluate_many(circuits: List[Circuit], settings: OptimizationSettings) -> List[torch.Tensor]: ...
def optimize_to_schedule(circuit: Circuit, settings: OptimizationSettings) -> Schedule: ...
def optimize_to_schedule_many(circuits: List[Circuit], settings: OptimizationSettings) -> Schedule: ...
def flat_concat(circuits: List[Circuit]) -> Concat: ...
def deep_heuristic_nest_adds(circuit: Circuit) -> Circuit: ...
def generalfunction_pull_concat(circuit: GeneralFunction) -> Optional[Concat]: ...
def concat_fuse(circuit: Concat) -> Optional[Concat]: ...
def concat_drop_size_zero(circuit: Concat) -> Optional[Concat]: ...
def index_concat_drop_unreached(circuit: Index) -> Optional[Index]: ...
def einsum_pull_concat(circuit: Einsum) -> Optional[Circuit]: ...
def add_pull_concat(circuit: Add) -> Optional[Circuit]: ...
def split_to_concat(circuit: Circuit, axis: int, sections: List[int]) -> Optional[Circuit]: ...
def deep_pull_concat_messy(circuit: Circuit, min_size: Optional[int]) -> Circuit: ...
def deep_pull_concat(circuit: Circuit, min_size: Optional[int]) -> Circuit: ...
def batch_inputs_axis_len(circuit: Circuit, axis_len: int, num_batches: int, min_size: Optional[int]) -> Circuit: ...
def batch_largest(circuit: Circuit, num_batches: int, min_size: Optional[int]) -> Circuit: ...
def set_named_axes(circuit: Circuit, named_axes: List[Optional[str]]) -> Circuit: ...
def propagate_named_axes(circuit: Circuit, named_axes: List[Optional[str]]) -> Circuit: ...
def toposort_circuit(circuit: Circuit) -> List[Circuit]: ...
def einsum_push_down_trace(circuit: Einsum) -> Optional[Einsum]: ...
def add_pull_diags(circuit: Add) -> Optional[Circuit]: ...
def concat_repeat_to_rearrange(circuit: Concat) -> Optional[Concat]: ...
def add_outer_product_broadcasts_on_top(circuit: Add) -> Optional[Add]: ...
def extract_add(add: Add, sub: Add) -> Optional[Add]: ...
def concat_to_scatter(concat: Concat) -> Optional[Scatter]: ...
def replace_all_randn_seeded(circuit: Circuit) -> Circuit: ...
def rust_expression_notation_circuit(circuit: Circuit) -> str: ...
def opt_eval_each_subcircuit_until_fail(circuit: Circuit, settings: OptimizationSettings) -> None: ...

# circuit manipulation functions
def deep_map(circuit: Circuit, fn: Callable[[Circuit], Circuit]) -> Circuit: ...
def deep_map_preorder(circuit: Circuit, fn: Callable[[Circuit], Circuit]) -> Circuit: ...
def filter_nodes(circuit: Circuit, fn: Callable[[Circuit], bool]) -> List[Circuit]: ...
def replace_nodes(circuit: Circuit, map: Dict[Circuit, Circuit]) -> Circuit: ...
def path_get(circuit: Circuit, path: List[int]) -> Circuit: ...
def update_nodes(
    circuit: Circuit, matcher: Callable[[Circuit], bool], mapper: Callable[[Circuit], Circuit]
) -> Circuit: ...
def update_path(circuit: Circuit, path: List[int], updater: Callable[[Circuit], Circuit]) -> Circuit: ...
